import unittest

from abst.sharing.local_broadcast import LocalBroadcast


class TestLocalBroadcast(unittest.TestCase):
    lb = None
    shared_memory_size = None
    shared_memory_name = None

    @classmethod
    def setUpClass(cls):
        # Use unique names for each test run to avoid conflicts
        cls.shared_memory_name = 'test_shared_memory'
        cls.shared_memory_size = 1024
        cls.lb = LocalBroadcast(cls.shared_memory_name, cls.shared_memory_size)

    @classmethod
    def tearDownClass(cls):
        # Cleanup shared memory
        cls.lb.close()

    def test_store_and_retrieve_json(self):
        test_data = {'key': 'value', 'integer': 42, 'float': 3.14, 'list': [1, 2, 3]}
        self.lb.store_json(test_data)
        retrieved_data = self.lb.retrieve_json()
        self.assertEqual(test_data, retrieved_data)

    def test_data_length_persistence(self):
        # This test ensures that the data length is correctly stored and retrieved,
        # implying it's shared correctly across instances/processes
        initial_data = {'key': 'initial'}
        size = self.lb.store_json(initial_data)

        # Create a new LocalBroadcast instance to mimic another process
        new_instance = LocalBroadcast(self.shared_memory_name, self.shared_memory_size)
        self.assertEqual(size, new_instance.get_used_space())
        initial_data_retrieved = new_instance.retrieve_json()
        self.assertEqual(initial_data, initial_data_retrieved)

        # Now store new data and verify the updated length is respected
        updated_data = {'key': 'updated', 'more': [1, 2, 3]}
        self.lb.store_json(updated_data)
        updated_data_retrieved = new_instance.retrieve_json()
        self.assertEqual(updated_data, updated_data_retrieved)

    def test_previous_data_not_overwritten(self):
        # Create a new LocalBroadcast instance to mimic another process
        new_instance_a = LocalBroadcast(self.shared_memory_name, self.shared_memory_size)
        new_instance_a.store_json({"Britney": "Hey"})
        new_instance_b = LocalBroadcast(self.shared_memory_name, self.shared_memory_size)
        new_instance_b.store_json({"Katty": "Whassup"})
        self.assertEqual(new_instance_b.retrieve_json(), {"Britney": "Hey", "Katty": "Whassup"})

    def test_previous_data_not_overwritten_same_key(self):
        # Create a new LocalBroadcast instance to mimic another process
        new_instance_a = LocalBroadcast(self.shared_memory_name, self.shared_memory_size)
        new_instance_a.store_json({"Britney": {"Katty": "Whassup"}})
        new_instance_b = LocalBroadcast(self.shared_memory_name, self.shared_memory_size)
        new_instance_b.store_json({"Britney": {"Chris": "Heey"}})
        self.assertEqual(new_instance_b.retrieve_json(), {"Britney": {"Katty": "Whassup", "Chris": "Heey"}})

    def test_exceed_memory_size(self):
        # This test verifies that an error is raised when trying to store data
        # that exceeds the allocated shared memory size
        large_data = {'key': 'x' * 2000}  # Assuming this exceeds the shared memory size
        with self.assertRaises(ValueError):
            self.lb.store_json(large_data)


if __name__ == '__main__':
    unittest.main()



import logging
import os
import re

import click
import rich
from InquirerPy import inquirer

from abst.utils.misc_funcs import setup_calls, fetch_pods


@click.group()
def pod():
    pass


@pod.command("ssh", help="Will SSH into pod with containing string name")
@click.argument("pod_name")
@click.option("--debug", is_flag=True, default=False)
def ssh_pod(pod_name, debug):
    setup_calls(debug)

    try:
        pod_lines = fetch_pods()
    except FileNotFoundError:
        rich.print("[red]kubectl not found on this machine[/red]")
        return

    found = list(filter(lambda pod_line: pod_name in pod_line, pod_lines))

    if len(found) > 1:
        data = re.sub(
            " +",
            " ",
            inquirer.select("Found more pods, choose one:", list(found)).execute(),
        ).split(" ")
        pod_name_precise = data[1]
    elif len(found) == 1:
        tmp = found.pop()
        data = re.sub(" +", " ", tmp).split(" ")
        pod_name_precise = data[1]
    else:
        rich.print(f"[red]No pods with name {pod_name} found[/red]")
        return

    rich.print(
        f"[green]Connecting to {pod_name_precise} in namespace: {data[0]}[/green]"
    )
    kubectl_ssh_cmd = f"kubectl exec -n {data[0]} --stdin --tty {pod_name_precise} -- /bin/bash"
    logging.info(f"Executing {kubectl_ssh_cmd}")
    os.system(
        kubectl_ssh_cmd
    )


@pod.command("logs", help="Will get logs from a pod with containing string name")
@click.argument("pod_name")
@click.option("--debug", is_flag=True, default=False)
def log_pod(pod_name, debug):
    setup_calls(debug)

    try:
        pod_lines = fetch_pods()
    except FileNotFoundError:
        rich.print("[red]kubectl not found on this machine[/red]")
        return

    found = list(filter(lambda pod_line: pod_name in pod_line, pod_lines))

    if len(found) > 1:
        data = re.sub(
            " +",
            " ",
            inquirer.select("Found more pods, choose one:", found).execute(),
        ).split(" ")
        pod_name_precise = data[1]
    elif len(found) == 1:
        tmp = found.pop()
        data = re.sub(" +", " ", tmp).split(" ")
        pod_name_precise = data[1]
    else:
        rich.print(f"[red]No pods with name {pod_name} found[/red]")
        return

    rich.print(
        f"[green]Getting logs from {pod_name_precise} in namespace: {data[0]}[/green]"
    )
    kubectl_log_cmd = f"kubectl -n {data[0]} logs {pod_name_precise}"
    logging.info(f"Executing {kubectl_log_cmd}")
    os.system(
        kubectl_log_cmd
    )



import logging
import os
import re
from pathlib import Path
from subprocess import call

import click
import rich
from InquirerPy import inquirer

from abst.utils.misc_funcs import setup_calls, fetch_pods, recursive_copy, copy_file_alt_kubectl


@click.group("cp", help="Copy commands for special usage")
def cp():
    pass


@cp.command("login", help="Login to docker and helm registry")
@click.argument("secret_name")
@click.argument("source_namespace")
@click.argument("target_namespace")
@click.option("--debug", is_flag=True, default=False)
def cp_secret(
        secret_name: str,
        target_namespace: str,
        source_namespace: str = "default",
        debug=False,
):
    """
    Copy Secret in current cluster from source namespace to target
    @param secret_name: Secret Name
    @param target_namespace: Target Namespace name
    @param source_namespace: Source Namespace name
    @return:
    :param debug:
    """
    setup_calls(debug)
    try:
        rich.print("Trying Copy secret")
        kubectl_copy_secret_cmd = f"kubectl get secret {secret_name} --namespace={source_namespace} -o yaml | sed " \
                                  f"'s/namespace: .*/namespace: {target_namespace}/' | kubectl apply -f -"
        logging.info(f"Executing {kubectl_copy_secret_cmd}")

        os.system(
            kubectl_copy_secret_cmd
        )
    except FileNotFoundError:
        rich.print("[red]kubectl not found on this machine[/red]")
        return


@cp.command("file", help="Will copy file into pod with containing string name")
@click.argument("pod_name", type=str)
@click.argument("local_path", type=str)
@click.argument("dest_path", type=str)
@click.option("--exclude", default="", type=str)
@click.option("--debug", is_flag=True, default=False)
def cp_to_pod(pod_name, local_path, dest_path, exclude, debug):
    setup_calls(debug)

    try:
        pod_lines = fetch_pods()
    except FileNotFoundError:
        rich.print("[red]kubectl not found on this machine[/red]")
        return

    found = list(filter(lambda pod_line: pod_name in pod_line, pod_lines))

    if len(found) > 1:
        data = re.sub(
            " +",
            " ",
            inquirer.select("Found more pods, choose one:", list(found)).execute(),
        ).split(" ")
        pod_name_precise = data[1]
    elif len(found) == 1:
        tmp = found.pop()
        data = re.sub(" +", " ", tmp).split(" ")
        pod_name_precise = data[1]
    else:
        rich.print(f"[red]No pods with name {pod_name} found[/red]")
        return

    rich.print(
        f"[green]Copying file {local_path} to {pod_name_precise}:{dest_path} in namespace: {data[0]}[/green]"
    )
    kubectl_copy_cmd = f"kubectl cp {local_path} {data[0]}/{pod_name_precise}:{dest_path}"
    logging.info(f"Executing {kubectl_copy_cmd}")

    exit_code = call(
        kubectl_copy_cmd.split(" ")
    )

    if exit_code == 255:
        rich.print("[red]Failed to copy using conventional kubectl cp, probably missing [yellow]tar[/yellow] "
                   "executable[/red]")
        rich.print("[yellow]Trying alternative copy method...[/yellow]")
    else:
        return
    local_path_obj = Path(local_path).expanduser().resolve()
    thread_list = []

    if not local_path_obj.is_dir():
        copy_file_alt_kubectl(data, dest_path, local_path, pod_name_precise)
        return

    recursive_copy(local_path_obj, dest_path, exclude, data, pod_name_precise, thread_list)

    for t in thread_list:
        t.join()



import json
import struct
from json import JSONDecodeError
from multiprocessing import shared_memory

from deepmerge import always_merger

from abst.config import max_json_shared


class LocalBroadcast:
    _instance = None

    def __new__(cls, name: str, size: int = max_json_shared):
        if cls._instance is None:
            cls._instance = super(LocalBroadcast, cls).__new__(cls)
            cls._instance.__init_shared_memory(name[:14], size)
        return cls._instance

    def __init_shared_memory(self, name: str, size: int):
        self._data_name = name
        self._len_name = f"{name}_len"
        self._size = size

        try:
            # Attempt to create the main shared memory block
            self._data_shm = shared_memory.SharedMemory(name=self._data_name, create=True, size=size)
            self._data_is_owner = True
        except FileExistsError:
            self._data_shm = shared_memory.SharedMemory(name=self._data_name, create=False)
            self._data_is_owner = False

        try:
            self._len_shm = shared_memory.SharedMemory(name=self._len_name, create=True, size=8)
            self._len_shm.buf[:8] = struct.pack('Q', 0)
            self._len_is_owner = True
        except FileExistsError:
            self._len_shm = shared_memory.SharedMemory(name=self._len_name, create=False)
            self._len_is_owner = False

    def store_json(self, data: dict) -> int:
        """
        Serialize and store JSON data in shared memory.
        @return: Size of the serialized data in bytes
        """

        data_before = self.retrieve_json()
        for key, value in data.items():
            for s_key in value.keys():
                if type(data_before[key][s_key]) == type(data[key][s_key]):
                    data_before[key].pop(s_key)

        data_copy = always_merger.merge(data, data_before)

        serialized_data = json.dumps(data_copy).encode('utf-8')
        if len(serialized_data) > self._size:
            raise ValueError("Data exceeds allocated shared memory size.")

        # Write the data length to the length shared memory
        self._len_shm.buf[:8] = struct.pack('Q', len(serialized_data))

        # Write data to the main shared memory
        self._data_shm.buf[:len(serialized_data)] = serialized_data
        return len(serialized_data)

    def retrieve_json(self) -> dict:
        """
        Retrieve and deserialize JSON data from shared memory.
        """
        # Read the data length from the length shared memory
        data_length = self.get_used_space()

        if data_length == -1:
            return {}

        # Read data from the main shared memory
        data = bytes(self._data_shm.buf[:data_length]).decode('utf-8')
        try:
            return json.loads(data)
        except JSONDecodeError:
            return {}

    def get_used_space(self) -> int:
        """
        Get the size of the shared memory
        @return: Number of bytes used
        """
        if self._len_shm.buf is None:
            return -1
        return struct.unpack('Q', self._len_shm.buf[:8])[0]

    def close(self):
        """Close and unlink the shared memory blocks."""
        try:
            self._data_shm.close()
            self._len_shm.close()
            if self._data_is_owner:
                self._data_shm.unlink()
            if self._len_is_owner:
                self._len_shm.unlink()
        except FileNotFoundError:
            pass



import click
import rich
from InquirerPy import inquirer

from abst.cli_commands.ssh_cli.utils import filter_keys_by_substring, filter_keys_by_port, do_ssh
from abst.config import broadcast_shm_name
from abst.sharing.local_broadcast import LocalBroadcast
from abst.utils.misc_funcs import setup_calls


@click.command("ssh", help="Will SSH into pod with containing string name")
@click.argument("port", required=False, default=None)
@click.option("-n", "--name", default=None, help="Name of context running, it can be just part of the name")
@click.option("--debug", is_flag=True, default=False)
def ssh_lin(port, name, debug):
    setup_calls(debug)

    lb = LocalBroadcast(broadcast_shm_name)
    data = lb.retrieve_json()

    if name and len(keys := filter_keys_by_substring(data, name)) == 1:
        do_ssh(keys[0], data[keys[0]]["username"], data[keys[0]]["port"])
        return

    if port is not None and len(keys := filter_keys_by_port(data, port)) == 1:
        do_ssh(keys[0], data[keys[0]]["username"], data[keys[0]]["port"])
        return

    longest_key = max(len(key) for key in data.keys())
    questions = [{"name": f"{key.ljust(longest_key)} |= status: {data['status']}", "value": (key, data)} for key, data
                 in data.items()]

    context_name, context = inquirer.select("Select context to ssh to:", questions).execute()

    if "username" not in context:
        rich.print("[red]Please fill resource-os-username into config for this feature to work[/red]")
        return
    elif "port" not in context:
        rich.print("[red]Please fill local-port into config for this feature to work[/red]")
        return

    do_ssh(name, context["username"], context["port"])
